from __future__ import annotations
from collections import defaultdict
from typing import Dict, List, Tuple
import math

def compute_flakiness(raw_rows: List[Tuple[str, str]]) -> Dict[str, Dict]:
    """
    raw_rows: list of (full_name, status) across many runs (mixed order).
    Return dict: {full_name: {pass_count, fail_count, transitions, flake_rate, suspect_flaky:bool}}
    """
    # Group by test
    history: Dict[str, List[str]] = defaultdict(list)
    for full_name, status in raw_rows:
        history[full_name].append(status)

    out: Dict[str, Dict] = {}
    for name, statuses in history.items():
        pass_count = sum(1 for s in statuses if s == "pass")
        fail_like = sum(1 for s in statuses if s in ("fail","error"))
        total = len(statuses)
        # transitions
        transitions = 0
        for i in range(1, len(statuses)):
            if statuses[i] != statuses[i-1]:
                transitions += 1
        flake_rate = fail_like / total if total else 0
        
        # Statistical significance scoring
        confidence_score = calculate_flakiness_confidence(pass_count, fail_like, total, transitions)
        
        # Updated logic for better classification:
        # For mixed results (potentially flaky tests)
        if pass_count > 0 and fail_like > 0:
            # Mixed results - use confidence threshold
            truly_flaky = confidence_score >= 0.7  # High confidence required for flaky classification
            always_failing = False
        else:
            # Consistent behavior
            truly_flaky = False
            always_failing = fail_like > 0 and pass_count == 0 and total >= 2  # Only failures, minimum runs
        
        # A test is "suspect" if it needs attention (either flaky or always failing)
        suspect = truly_flaky or always_failing
        
        out[name] = {
            "pass_count": pass_count,
            "fail_count": fail_like,
            "total_runs": total,
            "transitions": transitions,
            "flake_rate": flake_rate,
            "suspect_flaky": suspect,
            "truly_flaky": truly_flaky,
            "always_failing": always_failing,
            "confidence_score": confidence_score,
        }
    return out

def calculate_flakiness_confidence(pass_count: int, fail_count: int, total: int, transitions: int) -> float:
    """
    Calculate confidence in our flakiness assessment (0.0 to 1.0)
    - For consistent behavior (all pass/all fail): HIGH confidence in stability
    - For mixed results: confidence based on statistical significance of flakiness
    
    Returns 0.0 to 1.0 where 1.0 is high confidence in our assessment
    """
    if total < 1:
        return 0.0  # insufficient data
    
    # Base confidence on sample size (more runs = higher confidence in our assessment)
    sample_size_factor = min(1.0, total / 10.0)  # plateau at 10 runs for practical purposes
    
    # Case 1: Consistent behavior (all pass OR all fail)
    if pass_count == 0 or fail_count == 0:
        # High confidence that this test is STABLE (not flaky)
        # More runs = higher confidence in this assessment
        return min(0.95, 0.5 + sample_size_factor * 0.45)  # Start at 50%, max 95%
    
    # Case 2: Mixed results - potentially flaky
    # Calculate confidence based on how "flaky" the pattern looks
    
    # Transition rate (true flakiness shows state changes)
    transition_rate = transitions / max(1, total - 1) if total > 1 else 0
    
    # Balanced failure rate (around 20-80% suggests true flakiness)
    fail_rate = fail_count / total
    balance_factor = 1.0 - abs(0.5 - fail_rate) * 2  # Peaks at 50% fail rate
    
    # Statistical significance using Wilson score interval
    p = fail_rate
    z = 1.96  # 95% confidence
    if total > 1:
        denominator = 1 + z**2 / total
        center = (p + z**2 / (2 * total)) / denominator
        margin = z * math.sqrt((p * (1 - p) + z**2 / (4 * total)) / total) / denominator
        interval_width = 2 * margin
        # Narrower interval = higher confidence
        statistical_factor = max(0.0, 1.0 - interval_width)
    else:
        statistical_factor = 0.0
    
    # Combine all factors for mixed results
    flaky_confidence = (
        sample_size_factor * 0.4 +      # 40% weight on sample size
        transition_rate * 0.3 +         # 30% weight on transitions  
        balance_factor * 0.2 +          # 20% weight on balanced failure rate
        statistical_factor * 0.1        # 10% weight on statistical significance
    )
    
    return min(1.0, max(0.1, flaky_confidence))  # Minimum 10% for mixed results